#include "SkillManagerComponentPlugin.h"

#include <ArmarXCore/core/Component.h>
#include "../error/Exception.h"

namespace armarx::plugins
{
    void SkillManagerComponentPlugin::preOnInitComponent()
    {}

    void SkillManagerComponentPlugin::preOnConnectComponent()
    {}

    void SkillManagerComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {}
}


namespace armarx
{
    SkillManagerComponentPluginUser::SkillManagerComponentPluginUser()
    {
        addPlugin(plugin);
    }

    void SkillManagerComponentPluginUser::addProvider(const skills::manager::dto::ProviderInfo& info, const Ice::Current&)
    {
        std::lock_guard l(skillProviderMapMutex);
        if (skillProviderMap.find(info.providerName) == skillProviderMap.end())
        {
            ARMARX_INFO << "Adding a provider with name '" << info.providerName << "'.";
            skillProviderMap.insert({info.providerName, info.provider});
        }
        else
        {
            ARMARX_INFO << "Trying to add a provider with name '" << info.providerName << "' but the provider already exists. "
                        << "Overwriting the old provider info.";
            skillProviderMap[info.providerName] = info.provider;
        }
    }

    void SkillManagerComponentPluginUser::removeProvider(const std::string& providerName, const Ice::Current&)
    {
        std::lock_guard l(skillProviderMapMutex);
        if (auto it = skillProviderMap.find(providerName); it != skillProviderMap.end())
        {
            ARMARX_INFO << "Removing a provider with name '" << providerName << "'.";
            skillProviderMap.erase(it);
        }
        else
        {
            ARMARX_INFO << "Trying to remove a provider with name '" << providerName << "' but it couldn't be found.";
        }
    }

    std::string SkillManagerComponentPluginUser::getFirstProviderNameThatHasSkill(const std::string& skillName)
    {
        for (const auto& [providerName, providerPrx] : skillProviderMap)
        {
            auto allSkills = providerPrx->getSkillDescriptions();
            for (const auto& [currentSkillName, skillDesc] : allSkills)
            {
                if (currentSkillName == skillName)
                {
                    return providerName;
                }
            }
        }
        return "INVALID PROVIDER NAME";
    }

    skills::provider::dto::SkillStatusUpdate SkillManagerComponentPluginUser::executeSkill(const skills::manager::dto::SkillExecutionRequest& info, const Ice::Current&)
    {
        std::string providerName = "INVALID PROVIDER NAME";
        if (info.skillId.providerName == "*")
        {
            providerName = getFirstProviderNameThatHasSkill(info.skillId.skillName);
        }
        else if(not(info.skillId.providerName.empty()))
        {
            providerName = info.skillId.providerName;
        }

        bool remove = false;
        if (auto it = skillProviderMap.find(providerName); it != skillProviderMap.end())
        {
            const auto& n = it->first;
            const auto& s = it->second;

            if (s)
            {
                skills::callback::dti::SkillProviderCallbackInterfacePrx myPrx;
                getProxy(myPrx, -1);

                skills::provider::dto::SkillExecutionRequest exInfo;
                exInfo.skillName = info.skillId.skillName;
                exInfo.executorName = info.executorName;
                exInfo.callbackInterface = myPrx;
                exInfo.params = info.params;

                return s->executeSkill(exInfo);
            }
            else
            {
                remove = true;
            }

            if (remove)
            {
                std::scoped_lock l(skillProviderMapMutex);
                if (auto it = skillProviderMap.find(providerName); it != skillProviderMap.end())
                {
                    ARMARX_WARNING << __PRETTY_FUNCTION__ << ": Found disconnected skill provider '" << n << "' during execution. Removing it from skills.";
                    it = skillProviderMap.erase(it);
                }
            }
        }
        else
        {
            ARMARX_ERROR << "Could not execute a skill of provider '" + providerName + "' because the provider does not exist.";
            throw skills::error::SkillException(__PRETTY_FUNCTION__, "Skill execution failed. Could not execute a skill of provider '" + providerName + "' because the provider does not exist.");
        }
        return {}; // Never happens
    }

    void SkillManagerComponentPluginUser::abortSkill(const std::string& providerName, const std::string& skillName, const Ice::Current &current)
    {
        std::scoped_lock l(skillProviderMapMutex);
        if (auto it = skillProviderMap.find(providerName); it != skillProviderMap.end())
        {
            const auto& n = it->first;
            const auto& s = it->second;
            if (s)
            {
                s->abortSkill(skillName);
            }
            else
            {
                ARMARX_WARNING << __PRETTY_FUNCTION__ << ": Found disconnected skill provider '" << n << "'. Removing it from skills.";
                it = skillProviderMap.erase(it);
            }
        }
    }

    void SkillManagerComponentPluginUser::updateStatusForSkill(const skills::provider::dto::SkillStatusUpdate& statusUpdate, const Ice::Current&)
    {
        (void) statusUpdate;
        // If you want to use the status, implement this method!
    }

    skills::manager::dto::SkillDescriptionMapMap SkillManagerComponentPluginUser::getSkillDescriptions(const Ice::Current &current)
    {
        skills::manager::dto::SkillDescriptionMapMap ret;

        std::scoped_lock l(skillProviderMapMutex);
        for (auto it = skillProviderMap.cbegin(); it != skillProviderMap.cend();)
        {
            const auto& n = it->first;
            const auto& s = it->second;
            if (s)
            {
                skills::provider::dto::SkillDescriptionMap m = s->getSkillDescriptions();
                ret.insert({n, m});
                ++it;
            }
            else
            {
                ARMARX_WARNING << __PRETTY_FUNCTION__ << ": Found disconnected skill provider '" << n << "'. Removing it from skills.";
                it = skillProviderMap.erase(it);
            }
        }
        return ret;
    }

    skills::manager::dto::SkillStatusUpdateMapMap SkillManagerComponentPluginUser::getSkillExecutionStatuses(const Ice::Current &current)
    {
        skills::manager::dto::SkillStatusUpdateMapMap ret;

        std::scoped_lock l(skillProviderMapMutex);
        for (auto it = skillProviderMap.cbegin(); it != skillProviderMap.cend();)
        {
            const auto& n = it->first;
            const auto& s = it->second;
            if (s)
            {
                skills::provider::dto::SkillStatusUpdateMap m = s->getSkillExecutionStatuses();
                ret.insert({n, m});
                it++;
            }
            else
            {
                ARMARX_WARNING << __PRETTY_FUNCTION__ << ": Found disconnected skill provider '" << n << "'. Removing it from skills.";
                it = skillProviderMap.erase(it);
            }
        }
        return ret;
    }
}
